#include <Drivers.h>
#include <KernelExport.h>
#include <sys/param.h>
#include <OS.h>
#include <unistd.h>
#include <stdlib.h>

#include "mmap.h"

const char * device_names[]={"misc/mmap",NULL};
extern device_hooks mmap_hooks;

typedef struct {
	area_id kern_area;
	area_id user_area;
	void*area_address;
	off_t offset;
	ssize_t len;
	int fd;
	sem_id sem;
} cookie_t;

status_t init_hardware(void) { dprintf("mmap init_hardware\n"); {
	return B_OK;
}}

status_t init_driver(void) { dprintf("mmap init_driver\n"); {
	return B_OK;
}}

void uninit_driver(void) { dprintf("mmap uninit_driver\n"); {
}}

const char * * publish_devices() { dprintf("mmap publish_devices\n"); {
	return device_names;
}}

device_hooks * find_device(const char*name) { (void)name;dprintf("mmap find_device\n"); {
	return &mmap_hooks;
}}

status_t mmap_map(cookie_t*cookie,mmap_params_t*params) { dprintf("mmap_map\n"); {
	size_t area_size;

	if (cookie->fd!=-1) {
		dprintf("mmap_map : cannot map two files with the same mmap instance\n");
		return B_ERROR;
	}

	area_size=(params->len+B_PAGE_SIZE-1)&-B_PAGE_SIZE;

	dprintf("mmap_map : filename : %s\n",params->filename);
	dprintf("mmap_map : offset : %Ld\n",params->offset);
	dprintf("mmap_map : len : %d\n",(int)params->len);
	dprintf("mmap_map : area len : %d\n",(int)area_size);

	cookie->fd=open(params->filename,O_RDWR);
	cookie->offset=params->offset;
	cookie->len=params->len;
	if (cookie->fd==-1) {
		dprintf("mmap_map error in open()\n");
		goto err_file;
	}
	dprintf("mmap_map : open()ed fd %d\n",cookie->fd);

	dprintf("mmap_map : will create area\n");
	cookie->kern_area=create_area("mmap_kernel",&(cookie->area_address),B_ANY_KERNEL_ADDRESS,area_size,B_NO_LOCK,B_READ_AREA|B_WRITE_AREA);
	if (cookie->kern_area<0) {
		dprintf("mmap_map : error creating area\n");
		goto err_area;
	}
	dprintf("mmap_map : create_area()ed id %d, addr %p\n",(int)cookie->kern_area,cookie->area_address);

	dprintf("mmap_map : will clone area\n");
	cookie->user_area=clone_area("mmap_user",&(params->address),B_ANY_ADDRESS,B_READ_AREA|B_WRITE_AREA,cookie->kern_area);
	if (cookie->user_area<0) {
		dprintf("mmap_map : error cloning area\n");
		goto err_clone;
	}
	dprintf("mmap_map : clone_area()ed id %d, addr %p\n",(int)cookie->user_area,params->address);

	dprintf("mmap_map : will read from the file\n");
	if (lseek(cookie->fd,cookie->offset,SEEK_SET)==-1) {
		dprintf("mmap_map : error in lseek\n");
		goto err_read;
	}
	if (read(cookie->fd,cookie->area_address,cookie->len)!=cookie->len) {
		dprintf("mmap_map : error in read\n");
		goto err_read;
	}

	return B_OK;

err_read:
	dprintf("mmap_map : cleanup : will delete user area\n");
	if (delete_area(cookie->user_area)!=B_OK) {
		dprintf("mmap_map : error deleting user area\n");
	}
	cookie->user_area=0;
err_clone:
	dprintf("mmap_map : cleanup : will delete kernel area\n");
	if (delete_area(cookie->kern_area)!=B_OK) {
		dprintf("mmap_map : error deleting kernel area\n");
	}
	cookie->kern_area=0;
err_area:
	dprintf("mmap_map : cleanup : will close file\n");
	if (close(cookie->fd)!=0) {
		dprintf("mmap_map : error closing file\n");
	}
	cookie->fd=-1;
err_file:
	return B_ERROR;
}}

status_t mmap_wback(cookie_t*cookie) { dprintf("mmap_wback\n"); {
	if (cookie->fd==-1) {
		dprintf("mmap_wback : error : no mapped file\n");
		return B_ERROR;
	} else {
		dprintf("mmap_wback : will write to the file\n");
		if (lseek(cookie->fd,cookie->offset,SEEK_SET)==-1) {
			dprintf("mmap_wback : error in lseek\n");
			return B_ERROR;
		} else {
			if (write(cookie->fd,cookie->area_address,cookie->len)!=cookie->len) {
				dprintf("mmap_wback : error in write\n");
				return B_ERROR;
			}
		}
	}
	return B_OK;
}}

status_t mmap_unmap(cookie_t*cookie) { dprintf("mmap_unmap\n"); {
	status_t ret;
	ret=B_OK;
	if (cookie->fd==-1) {
		dprintf("mmap_unmap : error : no mapped file\n");
		return B_ERROR;
	} else {
		dprintf("mmap_unmap : will write data back\n");
		if (mmap_wback(cookie)!=B_OK) {
			dprintf("mmap_unmap : error writing data back\n");
		}
	
		dprintf("mmap_unmap : will close file %d\n",cookie->fd);
		if (close(cookie->fd)!=0) {
			dprintf("mmap_unmap : error closing file\n");
			ret=B_ERROR;
		}
		cookie->fd=-1;

		dprintf("mmap_unmap : will delete kernel area %d\n",(int)cookie->kern_area);
		if (delete_area(cookie->kern_area)!=B_OK) {
			dprintf("mmap_unmap : error deleting kernel area\n");
			ret=B_ERROR;
		}
		cookie->kern_area=0;

		dprintf("mmap_unmap : will delete user area %d\n",(int)cookie->user_area);
		if (delete_area(cookie->user_area)!=B_OK) {
			dprintf("mmap_unmap : error deleting user area\n");
			ret=B_ERROR;
		}
		cookie->user_area=0;
	}
	return ret;
}}

status_t mmap_open(const char*name,uint32 flags,cookie_t**cookie) { (void)name;(void)flags;dprintf("mmap_open\n"); {
	dprintf("mmap_open : will allocate cookie\n");
	*cookie=(void*)malloc(sizeof(cookie_t));
	if (*cookie==NULL) {
		dprintf("mmap_open : error allocating cookie\n");
		goto err0;
	}
	memset(*cookie,0,sizeof(cookie_t));
	(*cookie)->fd=-1;
	dprintf("mmap_open : will create semaphore\n");
	if (((*cookie)->sem=create_sem(1,"mmap_ioctl_lock"))<B_OK) {
		dprintf("mmap_open : error creating semaphore\n");
		goto err1;
	}
	return B_OK;
err1:
	dprintf("mmap_open : cleanup : will free cookie\n");
	free(*cookie);
	*cookie=NULL;
err0:
	return B_ERROR;
}}

status_t mmap_close(void*cookie) { (void)cookie;dprintf("mmap_close\n"); {
	return B_OK;
}}

status_t mmap_free(cookie_t*cookie) { dprintf("mmap_free\n"); {
	status_t err;
	if (cookie->fd!=-1) {
		dprintf("mmap_free : will unmap file\n");
		if (mmap_unmap(cookie)!=B_OK) {
			dprintf("mmap_free : error unmapping file\n");
			err=B_ERROR;
		}
	}
	dprintf("mmap_free : will delete semaphore %d\n",(int)cookie->sem);
	if (delete_sem(cookie->sem)!=B_OK) {
		dprintf("mmap_free : error deleting semaphore\n");
		err=B_ERROR;
	}
	dprintf("mmap_free : will free cookie\n");
	free(cookie);
	return B_OK;
}}

status_t mmap_ioctl(cookie_t*cookie,uint32 op,void*data,size_t len) { (void)len;dprintf("mmap_ioctl\n"); {
	status_t ret;
	do {
		ret=acquire_sem_etc(cookie->sem,1,B_CAN_INTERRUPT,0);
	} while(ret==B_INTERRUPTED);
	if (ret!=B_OK) {
		dprintf("mmap_ioctl : error acquiring semaphore\n");
		return ret;
	} else {
		switch(op) {
			case IOCTL_MMAP_MAP : {
				ret=mmap_map(cookie,data);
				break;
			}
			case IOCTL_MMAP_WBACK : {
				ret=mmap_wback(cookie);
				break;
			}
			case IOCTL_MMAP_UNMAP : {
				ret=mmap_unmap(cookie);
				break;
			}
			default : {
				ret=B_BAD_VALUE;
				break;
			}
		}
	}
	if (release_sem(cookie->sem)!=B_OK) {
		dprintf("mmap_ioctl : error releasing semaphore\n");
		ret=B_ERROR;
	}
	return ret;
}}

status_t mmap_read(void*cookie,off_t position,void*data,size_t*numbytes) { (void)cookie;(void)position;(void)data;dprintf("mmap_read\n"); {
	*numbytes=0;
	return B_OK;
}}

status_t mmap_write(void*cookie,off_t position,const void*data,size_t*numbytes) { (void)cookie;(void)position;(void)data;dprintf("mmap_write\n"); {
	*numbytes=0;
	return B_OK;
}}

device_hooks mmap_hooks={
	(device_open_hook)mmap_open,
	mmap_close,
	(device_free_hook)mmap_free,
	(device_control_hook)mmap_ioctl,
	mmap_read,
	mmap_write,
	NULL,
	NULL,
	NULL,
	NULL
};
